#!/usr/bin/env python
## gportage version 0.1 (20 Aug 2003,
## This is the first version with a version number)
## A simple portage package browser for PyGTK 2
##
## Copyright (c) 2003, Fredrik Arnerup (e97_far@e.kth.se)
## All rights reserved.
##
## Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##   * Redistributions of source code must retain the above copyright notice,
##     this list of conditions and the following disclaimer.
##
##   * Redistributions in binary form must reproduce the above copyright
##     notice, this list of conditions and the following disclaimer in the
##     documentation and/or other materials provided with the distribution.
##
## THIS SOFTWARE IS PROVIDED BY FREDRIK ARNERUP "AS IS" AND ANY
## EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
## WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
## DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
## FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
## DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
## SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
## CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
## LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
## OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
## DAMAGE.

import sys, pygtk
pygtk.require('2.0')
try:
    import portage
except ImportError:
    sys.exit("Could not find portage module.\n"
             "Are you sure this is a Gentoo system?")
import gtk, gobject, re, string, threading, time, os
gtk.threads_init()  # make sure gtk lets other threads run too

debug = 0
browser = 'gnome-moz-remote'

def get_version(ebuild):
    """Extract version number from ebuild name"""
    result = ''
    parts = portage.catpkgsplit(ebuild)
    if parts:
        result = parts[2]
        if parts[3] != 'r0':
            result += '-' + parts[3]
    return result


class PackageData:
    """An entry in the package database"""

    def __init__(self, fullname):
        self.fullname = fullname
        self.desc = ''
        self.installed = portage.db["/"]["vartree"].dep_match(fullname)
        #self.read_description()  # too slow, no dough
        
    def get_name(self):
        return self.fullname.split('/')[1]

    def get_section(self):
        return self.fullname.split('/')[0]

    def get_latest_ebuild(self, include_masked=1):
        if include_masked:
            criterion = "match-all"
        else:
            criterion = "match-visible"
        return portage.best(portage.portdb.xmatch(criterion, self.fullname))

    def get_homepage(self):
        try:
            return portage.portdb.aux_get(self.get_latest_ebuild(),
                                          ["HOMEPAGE"])[0]
        except:
            return ''

    def get_installed(self):
        return self.installed

    def read_description(self):
        try:
            self.desc = portage.portdb.aux_get(
                self.get_latest_ebuild(),
                ["DESCRIPTION"])[0].encode('UTF-8')
        except Exception, e:
            self.desc = "An error occured when reading the description:\n" \
                        + str(e)


def sort(list):
    """sort in alphabetic instead of ASCIIbetic order"""
    spam = [(x[0].upper(), x) for x in list]
    spam.sort()
    return [x[1] for x in spam]


class ReadDatabaseThread(threading.Thread):

    def __init__(self):
        threading.Thread.__init__(self)
        self.done = 0
        self.count = 0
        self.error = ""
        
    def get_entry(self, fullname):
        data = PackageData(fullname)
        section = data.get_section()
        name = data.get_name()
        if section in self.db and name in self.db[section]:
            return self.db[section][name]
        else:
            return None

    def read_db(self):
        self.db = {}
        self.list = []
        tree = portage.db['/']['porttree']
        try:
            allnodes = tree.getallnodes()
        except OSError, e:
            # I once forgot to give read permissions
            # to an ebuild I created in the portage overlay
            self.error = str(e)
            return
        for entry in allnodes:
            if debug:
                print 'patang'
            self.count += 1
            section, name = entry.split('/')
            if not section in self.db:
                self.db[section] = {}
            data = PackageData(entry)
            self.db[section][name] = data;
            self.list.append((name, data))
        self.list = sort(self.list)

    def run(self):
        self.read_db()
        self.done = 1   # tell main thread that this thread has finished


def xml_esc(string):
    """Escape characters that have special meanings in XML"""
    def subst(c):
        if c == '<': return '&lt;'
        elif c == '>': return '&gt;'
        elif c == '&': return '&amp;'
        else: return c
    return ''.join(map(subst, string))


class Win:

    def open_webpage(self, button):
        # perhaps we should use the webbrowser module instead?
        os.system(browser + ' ' + button.url)

    def on_selection_changed(self, selection):
        name_markup = url = desc_markup = ''
        model, iter = selection.get_selected()
        if iter:
            data = model.get_value(iter, 0)
            self.current_data = data
            if not data: # a category
                name_markup = '<big>' + model.get_value(iter, 1) + '</big>'
            else:                
                # if it hasn't already been done:
                # read and cache data
                if not data.desc:
                    data.read_description()
                model.set_value(iter, 2,
                                "<i>" + xml_esc(data.desc) + "</i>")
                name_markup = '<big>' + xml_esc(data.fullname) + '</big>';
                url = data.get_homepage()
                desc_markup = '\n<b>' + xml_esc(data.desc) + '</b>\n\n'
                if data.installed:
                    desc_markup += \
                                "Latest version installed:\t" \
                                + get_version(portage.best(data.installed)) \
                                + "\n"
                else:
                    desc_markup += "Not installed\n"
                latest = get_version(
                    data.get_latest_ebuild(include_masked=1))
                latest_unmasked = get_version(
                    data.get_latest_ebuild(include_masked=0))
                if latest_unmasked:
                    desc_markup += "Latest unmasked version:\t" \
                              + latest_unmasked + "\n"
                if latest != latest_unmasked:
                    desc_markup += "Latest masked version:\t" + latest + "\n"
        for child in self.propbox.get_children():
            self.propbox.remove(child)
        label = gtk.Label(""); label.set_alignment(0, 0);
        label.set_markup(name_markup)
        self.propbox.pack_start(label, gtk.FALSE, gtk.FALSE)

        if url:
            button = gtk.Button(); button.set_relief(gtk.RELIEF_NONE)
            button.url = url  # save url in button
            button.connect("activate", self.open_webpage)
            button.connect("clicked", self.open_webpage)
            label = gtk.Label(""); label.set_alignment(0, 0);
            url_markup = '<u><span foreground="blue">' \
                         + xml_esc(url) + "</span></u>"
            label.set_markup(url_markup)
            button.add(label)
            self.propbox.pack_start(button, gtk.FALSE, gtk.FALSE)

        if desc_markup:
            label = gtk.Label(""); label.set_alignment(0, 0);
            label.set_line_wrap(gtk.TRUE);
            label.set_markup(desc_markup)
            self.propbox.pack_start(label, gtk.FALSE, gtk.FALSE)

        self.propbox.show_all()
        
    def add_entry(self, model, parent, data):
        entry_iter = model.insert_before(parent, None)
        model.set_value(entry_iter, 0, data)
        spam = xml_esc(data.get_name())
        if data.installed:
            spam = "<b>" + spam + "</b>"
        model.set_value(entry_iter, 1, spam)
        model.set_value(entry_iter, 2,
                        "<i>" + xml_esc(data.desc) + "</i>")

    def build_lists(self):
        db = db_thread.db
        list = db_thread.list
        sections = db.keys()
        sections.sort()
        count = 0
        # Categories:
        for section in sections: 
            section_iter = self.models[0].insert_before(None, None)
            self.models[0].set_value(section_iter, 1,
                                     '<b>' + xml_esc(section) + "</b>")
            entries = db[section].keys()
            entries.sort()
            for entry in entries:
                count += 1
                self.add_entry(self.models[0], section_iter,
                               db[section][entry])
        self.messages[0] = self.messages[1] = (str(count) + ' packages in '
                                               + str(len(db)) + ' categories')
        self.cafe_opera.push(0, self.messages[0])
        # All:
        for entry in list:
            self.add_entry(self.models[1], None, entry[1])
        # Installed:
        installed = sort(portage.db["/"]["vartree"].getallnodes())
        last_section = None
        for entry in installed:
            section, package = entry.split('/')
            if section != last_section:
                last_section = section
                section_iter = self.models[2].insert_before(None, None)
                self.models[2].set_value(section_iter, 1,
                                         '<b>' + xml_esc(section) + "</b>")
            data = db_thread.get_entry(entry)
            if data:
                self.add_entry(self.models[2], section_iter, data)
            else:
                # if installed from PORTDIR_OVERLAY,
                # it won't be in the database
                # Todo: fix this
                iter = self.models[2].insert_before(section_iter, None)
                self.models[2].set_value(iter, 1, xml_esc(package))
        self.messages[2] = (str(len(installed)) + ' installed packages')
                
    def on_page_change(self, notebook, spam, num):
        self.on_selection_changed(self.lists[num].get_selection())
        self.cafe_opera.pop(0)
        self.cafe_opera.push(0, self.messages[num])
        
    def on_search(self, entry):
        regexp = self.use_regexp.get_active()
        search_term = entry.get_text()
        if not (db_thread.done and search_term):
            return
        if regexp: # compile pattern for efficiency
            re_object = re.compile(search_term, re.I)
        else: # search is case-insensitive
            search_term = search_term.upper()
        self.models[3].clear()
        self.notebook.set_current_page(3)
        self.cafe_opera.push(0, 'Searching ...')
        count = 0
        for entry in db_thread.list:
            if regexp:
                match = re_object.search(entry[0]) != None
            else:
                match = string.find(entry[0].upper(), search_term) != -1
            if match:
                count += 1
                self.add_entry(self.models[3], None, entry[1])
        self.cafe_opera.pop(0)
        if count > 1:
            result = str(count) + " matches found"
        elif count == 1:
            result = "1 match found"
        else:
            result = "No matches"
        self.cafe_opera.push(0, result)
        self.messages[3] = result
        
    def __init__(self):
        self.current_data = None
        self.messages = ['', '', '', '']

        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.window.set_title("Portage Browser")

        vbox = gtk.VBox()

        hbox = gtk.HBox(gtk.FALSE, 12)
        hbox.set_border_width(12)
        label = gtk.Label("_Search:")
        label.set_use_underline(gtk.TRUE)
        hbox.pack_start(label, gtk.FALSE, gtk.FALSE)
        self.search_entry = gtk.Entry()
        self.search_entry.connect("activate", self.on_search)
        hbox.pack_start(self.search_entry)
        label.set_mnemonic_widget(self.search_entry)
        self.use_regexp = gtk.CheckButton("Use _regexp")
        self.use_regexp.set_use_underline(gtk.TRUE)
        hbox.pack_start(self.use_regexp, gtk.FALSE, gtk.FALSE)
        vbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)
        
        paned = gtk.VPaned()
        vbox.pack_start(paned)

        tabs = ["_Categories", "_All", "_Installed", "S_earch results"]
        self.notebook = gtk.Notebook()
        self.lists = []
        self.models = []
        for i in range(4):
            renderer1 = gtk.CellRendererText()
            renderer2 = gtk.CellRendererText()
            column1 = gtk.TreeViewColumn("Name", renderer1, markup = 1)
            column2 = gtk.TreeViewColumn("Description", renderer2, markup = 2)
            self.models.append(gtk.TreeStore(gobject.TYPE_PYOBJECT,
                                             gobject.TYPE_STRING,
                                             gobject.TYPE_STRING))
            self.lists.append(gtk.TreeView(self.models[i]))
            self.lists[i].set_enable_search(gtk.TRUE)
            self.lists[i].append_column(column1)
            self.lists[i].append_column(column2)
            self.lists[i].set_headers_visible(gtk.FALSE)
            self.lists[i].get_selection().connect("changed",
                                                  self.on_selection_changed)
            scroller = gtk.ScrolledWindow()
            scroller.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            scroller.set_shadow_type(gtk.SHADOW_IN)
            scroller.add(self.lists[i])
            label = gtk.Label(tabs[i])
            label.set_use_underline(gtk.TRUE)
            self.notebook.append_page(scroller, label)
        paned.pack1(self.notebook)
        self.notebook.set_size_request(400, 300)
        self.notebook.connect("switch-page", self.on_page_change)

        self.propbox = gtk.VBox();
        self.propbox.set_border_width(12)
        scroller = gtk.ScrolledWindow()
        scroller.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scroller.add_with_viewport(self.propbox)
        scroller.set_size_request(400, 200)
        paned.pack2(scroller)
        self.cafe_opera = gtk.Statusbar()
        self.cafe_opera.push(0, "")  # so the stack isn't empty
        vbox.pack_start(self.cafe_opera, gtk.FALSE, gtk.FALSE)
                
        self.window.connect("destroy", self.on_destroy)
        self.window.add(vbox)
        self.window.show_all()
        gtk.timeout_add(100, self.check_db_thread)

    def check_db_thread(self):
        if debug:
            print 'ekki'
        self.cafe_opera.pop(0)
        self.cafe_opera.push(0, 'Reading ... ('
                             + str(db_thread.count) + ' packages)')
        if db_thread.done:
            db_thread.join()
            if db_thread.error:
                sys.exit(db_thread.error) # Todo: display error dialog
            self.build_lists()
            return gtk.FALSE  # disable this callback
        else:
            return gtk.TRUE

    def on_destroy(self, widget, data = None):
        gtk.main_quit()
   

w = Win()
db_thread = ReadDatabaseThread(); db_thread.start()
gtk.main()
